/*
 *  TOPPERS ATK2
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *      Automotive Kernel Version 2
 *
 *  Copyright (C) 2012-2013 by Center for Embedded Computing Systems
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2012-2013 by FUJISOFT INCORPORATED, JAPAN
 *  Copyright (C) 2012-2013 by FUJITSU VLSI LIMITED, JAPAN
 *  Copyright (C) 2012-2013 by NEC Communication Systems, Ltd., JAPAN
 *  Copyright (C) 2012-2013 by Panasonic Advanced Technology Development Co., Ltd., JAPAN
 *  Copyright (C) 2012-2013 by Renesas Electronics Corporation, JAPAN
 *  Copyright (C) 2012-2013 by Sunny Giken Inc., JAPAN
 *  Copyright (C) 2012-2013 by TOSHIBA CORPORATION, JAPAN
 *  Copyright (C) 2012-2013 by Witz Corporation, JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  $Id: v850e2_fx4.c 7408 2013-04-19 15:51:11Z honda $
 */
#include "kernel_impl.h"
#include "v850e2_fx4.h"
#include "Os.h"
#include "prc_sil.h"

/*******************************************************************************************/
/*  Outline      : Write protected register                                                */
/*  Argument     : Register address                                                        */
/*                 Register data                                                           */
/*                 Register No                                                             */
/*  Return value : 0: write success / 1: write error                                       */
/*  Description  : Write protected register                                                */
/*                                                                                         */
/*******************************************************************************************/
static uint32
write_protected_reg(uint32 addr, uint32 data, uint8 regno)
{
	uint32	reg_top = 0xff420000;
	uint32	reg_stat;
	uint8	wk;
	SIL_PRE_LOC;

	if (regno > 2) {
		return(UC_INVALIDPARAM);
	}

	switch (regno) {
	case  0:
		reg_top += 0x4000;      /* PROTCMD0 */
		break;
	case  1:
		reg_top += 0x8000;      /* PROTCMD1 */
		break;
	case  2:
		reg_top += 0x0300;      /* PROTCMD2 */
		break;
	}
	reg_stat = reg_top + 4;     /* PROTS0/PROTS1/PROTS2 */

	SIL_LOC_INT();
	sil_wrb_mem((void *) reg_top, 0xA5);

	sil_wrw_mem((void *) addr, data);
	sil_wrw_mem((void *) addr, ~data);
	sil_wrw_mem((void *) addr, data);
	SIL_UNL_INT();

	wk = sil_reb_mem((void *) reg_stat);
	wk &= 0x01;

	return((wk == 0) ? UC_SUCCESS : UC_PROTREGERROR);
} /* write_protected_reg */

/********************************************************************************************/
/*  Function Name : V850Drv_nop                                                             */
/*  Input         : none                                                                    */
/*  Output        : none                                                                    */
/*  Description   : nop command                                                             */
/********************************************************************************************/
static void
V850Drv_nop(void)
{
	Asm("nop");

} /* V850Drv_nop */

/*******************************************************************************************/
/*  Outline      : Sub Oscillator enable                                                   */
/*  Argument     : -                                                                       */
/*  Return value : 0: successfly set / 1: set error                                        */
/*  Description  : Sub Oscillator register setting                                         */
/*                                                                                         */
/*******************************************************************************************/
uint32
EnableSubOSC(void)
{
	uint32 ucret;

	sil_wrw_mem((void *) SOSCST, 0x02);      /* stabilization time -> 262ms */

	ucret = write_protected_reg(SOSCE, 0x01, PROT_SOSCE);  /* SubOSC start */

	if (ucret != UC_SUCCESS) return(ucret);

	while (((sil_rew_mem((void *) SOSCS)) & CLK_S_CLKSTAB) == 0) {     /* Wait stabilization */
		V850Drv_nop();
	}
	return(UC_SUCCESS);
} /* EnableSubOSC */

/*******************************************************************************************/
/*  Outline      : Main Oscillator enable                                                  */
/*  Argument     : Main Cscillator frequency(Hz)                                           */
/*  Return value : 0: successfly set / 1: set error                                        */
/*  Description  : Main Oscillator register setting                                        */
/*                                                                                         */
/*******************************************************************************************/
uint32
EnableMainOSC(uint32 clk_in)
{
	uint8	ampsel;
	uint32	ucret;

	if (clk_in == CLK_MHz(4)) {
		ampsel = 0x03;
	}
	else if (CLK_MHz(4) < clk_in && clk_in <= CLK_MHz(8)) {
		ampsel = 0x02;
	}
	else if (CLK_MHz(8) < clk_in && clk_in <= CLK_MHz(16)) {
		ampsel = 0x01;
	}
	else if (CLK_MHz(16) < clk_in && clk_in <= CLK_MHz(20)) {
		ampsel = 0x00;
	}
	else {
		return(UC_INVALIDPARAM);
	}

	sil_wrw_mem((void *) MOSCC, 0x00 | ampsel); /* Normal stabilization time mode */
	sil_wrw_mem((void *) MOSCST, 0x0F);         /* stabilization time -> Max */

	ucret = write_protected_reg(MOSCE, 0x01, PROT_MOSCE);   /* MainOSC start */

	if (ucret != UC_SUCCESS) return(ucret);

	while (((sil_rew_mem((void *) MOSCS)) & CLK_S_CLKSTAB) == 0) {     /* Wait stabilization */
		V850Drv_nop();
	}

	return(UC_SUCCESS);
} /* EnableMainOSC */

/*******************************************************************************************/
/*  Outline      : PLL enable                                                              */
/*  Argument     : PLL No                                                                  */
/*                 Multiplying rate                                                        */
/*                 Register data                                                           */
/*  Return value : 0: successfly set / 1: set error                                        */
/*  Description  : PLL register setting                                                    */
/*                                                                                         */
/*******************************************************************************************/
static uint32
EnablePLL(uint32 pllno, uint8 clk_mul, uint8 p_val)
{
	uint32 ucret;

	if (clk_mul > 50)
		return(UC_INVALIDPARAM);

	sil_wrw_mem((void *) PLLC(pllno), (0x800000 | (p_val << 8) | (clk_mul - 1)));   /* PLL Mode, P, Nr */
	sil_wrw_mem((void *) PLLST(pllno),  0x07);                                      /* stabilization time -> Max */

	ucret = write_protected_reg(PLLE(pllno), 0x01, PROT_PLLE);  /* PLL Start */

	if (ucret != UC_SUCCESS)
		return(ucret);

	while (((sil_rew_mem((void *) PLLS(pllno))) & CLK_S_CLKSTAB) == 0) {     /* Wait stabilization */
		/* V850Drv_nop(); */
	}

	return(UC_SUCCESS);
} /* EnablePLL */

/*******************************************************************************************/
/*  Outline      : Set PLL frequency                                                       */
/*  Argument     : PLL select No                                                           */
/*                 PLL frequency                                                           */
/*                 PLL frequency                                                           */
/*  Return value : 0: successfly set / 1: set error                                        */
/*  Description  : Set PLL register from frequency                                         */
/*                                                                                         */
/*******************************************************************************************/
uint32
SetPLL(uint32 pllno, uint32 mhz, uint32 *outclk)
{
	uint32 mul;

	if (mhz < 20) {
		return(UC_ERROR);
	}
	else if (mhz < 50) {
		mul = (mhz / MAINOSC_CLOCK) * 4;
		if (mul <= 5 && 51 <= mul)
			return(UC_ERROR);
		*outclk = MAINOSC_CLOCK * mul / 4;
		return(EnablePLL(pllno, mul, PDIV4R0_025TO050));
	}
	else if (mhz < 100) {
		mul = (mhz / MAINOSC_CLOCK) * 2;
		if (mul <= 5 && 51 <= mul)
			return(UC_ERROR);
		*outclk = MAINOSC_CLOCK * mul / 2;
		return(EnablePLL(pllno, mul, PDIV2R0_050TO100));
	}
	else if (mhz < 200) {
		mul = (mhz / MAINOSC_CLOCK) * 1;
		if (mul <= 5 && 51 <= mul)
			return(UC_ERROR);
		*outclk = MAINOSC_CLOCK * mul / 1;
		return(EnablePLL(pllno, mul, PDIV1R0_100TO200));
	}
	else if (mhz < 400) {
		mul = (mhz / MAINOSC_CLOCK) / 2;
		if (mul <= 5 && 51 <= mul)
			return(UC_ERROR);
		*outclk = MAINOSC_CLOCK * mul * 2;
		return(EnablePLL(pllno, mul, PDIV0R5_200TO400));
	}
	return(UC_ERROR);
} /* SetPLL */

/*******************************************************************************************/
/*  Outline      : Clock switcher setting                                                  */
/*  Argument     : Control reginster address                                               */
/*                 Status reginster address                                                */
/*                 Register No                                                             */
/*                 Select No                                                               */
/*  Return value : 0: successfly set / 1: set error                                        */
/*  Description  : Select clock source of CKSCLK_mn                                        */
/*                                                                                         */
/*******************************************************************************************/
uint32
set_clock_selection(uint32 control, uint32 status, uint8 regno, uint16 sel)
{
	uint32 ucret;

	ucret = write_protected_reg(control, sel << 1, regno);
	if (ucret != UC_SUCCESS) {
		return(ucret);
	}

	/* Wait for SelectEnable */
	while (((sil_rew_mem((void *) status)) & 0x01) == 0) {
		V850Drv_nop();
	}

	return(UC_SUCCESS);
} /* set_clock_selection */
